/* 	$Id: global.h,v 1.5 1997/06/28 22:17:57 pirx Exp $	 */

/*
 * Author: Achim Bangert (abangert@ix.urz.uni-heidelberg.de)
 */

#ifndef GLOBAL_H
#define GLOBAL_H

#include "readproc.h"

#define DEFAULT_REFRESH 5000
#define XtRField "Field"
#define NUM_FIELDS 21

typedef enum {
  P_UID = 0, P_PID, P_PPID, P_TTY, P_TTYL, P_TIME, P_CTIME, P_PRI, P_NI, 
  P_SIZE, P_RSS, P_SWAP, P_FLAGS, P_WCHAN, P_SWCHAN, P_STAT, P_USER, 
  P_COMMAND, P_COMMANDLINE, P_PCPU, P_PMEM, P_END
} field_id_t;

typedef struct {
  String fields_menu;
  String signal_menu;
  String sel_pid_title;
  String sel_tty_title;
  String sel_uid_title;
  String sel_stat_title;
  String sig_signal_error;
  String sig_process_error;
  String sig_permission_error;
  String sig_success;
  String prio_process_error;
  String prio_permission_error;
  String prio_access_error;
  String prio_success;
  int refresh_intervall;
} AppResources;

extern AppResources app_resources;

typedef struct field_info_s {
  int map[NUM_FIELDS];
  int shown_field[NUM_FIELDS];
  int num_fields;
  int num_shown;
  char header[2048];
} field_info_t, *field_info_ptr;

extern field_info_t fieldinfo;

extern char *field_names[];
extern char *field_headers[];

typedef struct int_list_s {
  int *list;
  int num_allocated;
  int num_used;
} int_list_t, *int_list_ptr;

typedef struct string_list_s {
  String *list;
  int num_allocated;
  int num_used;
} string_list_t, *string_list_ptr;

typedef struct pid_list_s {
  pid_t *list;
  int num_allocated;
} pid_list_t, *pid_list_ptr;

typedef struct dev_list_s {
  dev_t *list;
  int num_allocated;
} dev_list_t, *dev_list_ptr;

typedef struct uid_list_s {
  uid_t *list;
  int num_allocated;
  int num_used;
} uid_list_t, *uid_list_ptr;

typedef struct history_list_s {
  int pid;
  int time;
} history_list_t, *history_list_ptr;

#define PROC_FILL (PROC_FILLMEM | PROC_FILLCMD | PROC_FILLTTY | PROC_FILLUSR)

typedef struct process_info_s {
  proc_t **table;
  String *str;
  history_list_ptr histlist;
  int num_allocated;
  int num_used;
  int num_hist;
} process_info_t, *process_info_ptr;

extern process_info_t procinfo;

#define NUM_FILTERS 8

#define FILT_PID    0
#define FILT_SELECT 1
#define FILT_TTY    2
#define FILT_ANYTTY 3
#define FILT_UID    4
#define FILT_OWN    5
#define FILT_STAT   6
#define FILT_NONE   7

#define PROC_NONE   0x1
#define PROC_SELECT (PROC_PID | PROC_NONE)
#define PROC_OWN    (PROC_UID | PROC_NONE)

typedef struct filter_info_s {
  int proc_flag[NUM_FILTERS];
  pid_list_t pid;
  dev_list_t tty;
  uid_list_t uid;
  char stat[6];
  string_list_t str;
} filter_info_t, *filter_info_ptr;

extern filter_info_t filtinfo;

#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))

extern XtAppContext app_context;
extern Display *app_display;
extern Screen *app_screen;
extern Pixmap tick_bitmap;
extern XtIntervalId refresh_timer;
extern Widget app_shell;
extern Widget button_bar;
extern Widget proc_selector;

#endif
















