/* 	$Id: xfields.c,v 1.7 1997/07/01 08:25:58 pirx Exp $	 */

/*
 * Author: Achim Bangert (abangert@ix.urz.uni-heidelberg.de)
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#ifdef XAW3D
#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/Sme.h>
#else
#include <X11/Xaw/Box.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#endif
#include "Listsel.h"
#include "SmeToggle.h"

#include "global.h"
#include "xtop.h"
#include "xfields.h"

/* used in resource files */
char *field_names[] =
{
  "UID", "PID", "PPID", "TTY", "TTYL", "TIME", "CTIME", "PRI", "NI", 
  "SIZE", "RSS", "SWAP", "FLAGS", "WCHAN", "SWCHAN", "STAT", "USER", 
  "COMMAND", "COMMANDLINE", "PCPU", "PMEM" 
};

/* corresponding headers */
char *field_headers[] =
{
  " UID ", "  PID ", " PPID ", "TTY ", "TTY    ", 
  "  TIME ", "  TIME ", "PRI ", " NI ", " SIZE ", " RSS ", "SWAP ", 
  "   FLAGS ", "WCHAN     ", "WCHAN     ", "STAT ", "USER     ", 
  "COMMAND", "COMMAND", "%CPU ", "%MEM "
};

field_info_t fieldinfo;

Widget fields_button;
Widget fields_shell;
Widget fields_option[NUM_FIELDS];

static void update_fieldinfo(void)
/* update fieldinfo to show of currently set fields */
{
  field_info_ptr f = &fieldinfo;
  Boolean state;
  int i, j;

  *(f->header) = '\0';
  for(i = j = 0; i < f->num_fields; ++i)
    {
      XtVaGetValues(fields_option[i], XtNstate, &state, NULL);
      if(state)
	{
	  f->shown_field[j++] = f->map[i];
	  strcat(f->header, field_headers[f->map[i]]);
	}
    }
  f->num_shown = j;
}

static void fields_callback(Widget w, XtPointer client, XtPointer call)
/* toggle field's state in fieldinfo.shown_map */
{
  int reason = (int) call;

  if(reason == XstoRADIO_TURNOFF)
    /* ignore; wait for callback by radio sibling which is now set */
    return;
  update_fieldinfo();
  XtVaSetValues(proc_selector, 
		XtNavailableLabel, fieldinfo.header,
 		XtNselectedLabel, fieldinfo.header,
		NULL);
  XtRemoveTimeOut(refresh_timer);
  refresh_display(NULL, &refresh_timer);
}

static void resource2fields_menu(String resource)
/* build fields menu from application resource */
{
  field_info_ptr f = &fieldinfo;
  Arg arglist[5];
  char buffer[128];
  char label[32];
  char name[32];
  char shown;
  int count;
  field_id_t fid;
  int num_read, num_left;

  count = 0;
  XtSetArg(arglist[0], XtNleftMargin, 20);
  XtSetArg(arglist[1], XtNplacement, XtJustifyLeft);
  XtSetArg(arglist[2], XtNonBitmap, tick_bitmap);
  for(num_left = strlen(resource); num_left > 0; num_left -= num_read)
    {
      num_read = -1;
      sscanf(resource, "%30[^:\n]:%30[^:\n]:%c\n%n", label, name, &shown, 
	     &num_read);
      if(num_read == -1)
	{
	  String param = buffer;
	  Cardinal num_param = 1;

	  sprintf(buffer, "\"%s\" in line %d of", resource, count);
	  XtAppErrorMsg(app_context, "conversionFailed", "xtopFieldMenu", 
			"XtopError", "Conversion of entry %s fieldMenu " 
			"resource failed", &param, &num_param);
	}
      if((fid = string2field(name)) < P_END)
	{
	  f->map[count] = fid;
	  sprintf(buffer, "show%s", name);
	  XtSetArg(arglist[3], XtNlabel, label);
	  XtSetArg(arglist[4], XtNstate, (shown == 'x'));
	  fields_option[count] = 
	    XtCreateManagedWidget(buffer, smeToggleObjectClass, fields_shell,
				  arglist, 5);
	  XtAddCallback(fields_option[count], XtNcallback, 
			fields_callback, (XtPointer) count);
	  ++count;
	}
      else
	{
	  String param = name;
	  Cardinal num_param = 1;

	  XtAppWarningMsg(app_context, "unknownField", "xtopFieldMenu", 
			  "XtopError", "Unknown field %s ignored",
			  &param, &num_param);
	}
      resource += num_read;
      /* proceed to next line */
    }
  f->num_fields = count;
  update_fieldinfo();
}

void create_fields_menu(void)
/* create menu for selecting displayed fields */
{
  fields_button =
    XtCreateManagedWidget("fieldsButton", menuButtonWidgetClass,
			  button_bar, NULL, 0);
  fields_shell = 
    XtCreatePopupShell("fieldsShell", simpleMenuWidgetClass, fields_button,
		       NULL, 0);
  XtVaSetValues(fields_button, XtNmenuName, "fieldsShell", NULL);
  resource2fields_menu(app_resources.fields_menu);
}











