/* 	$Id: xfilter.c,v 1.8 1997/07/01 08:25:44 pirx Exp $	 */

/*
 * Author: Achim Bangert (abangert@ix.urz.uni-heidelberg.de)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <assert.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#ifdef XAW3D
#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/Sme.h>
#include <X11/Xaw3d/SmeBSB.h>
#include <X11/Xaw3d/SmeLine.h>
#else
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#endif
#include "Listsel.h"
#include "SmeToggle.h"

#include "global.h"
#include "xtop.h"
#include "xfilter.h"

static String stat_list[] = {
  "R running", "S sleeping", "D comatose", "Z zombie", 
  "T traced/stopped"
};

static String filter_option_name[] = {
  "filterByPID", "filterBySELECT",  "filterByTTY", "filterByANYTTY", 
  "filterByUID", "filterByOWN", "filterBySTAT", "filterByNONE"
};

static String select_option_name[] = {
  "selectByTTY", "selectByUID", "selectByOWN", "selectBySTAT"
};

filter_info_t filtinfo = {
  {PROC_PID, PROC_SELECT, PROC_TTY, PROC_ANYTTY, PROC_UID, PROC_OWN, 
   PROC_STAT, PROC_NONE}, /* proc_flag */
  {NULL, 0},              /* pid */
  {NULL, 0},              /* tty */
  {NULL, 0, 0},           /* uid */
  "\0",                   /* stat */
  {NULL, 0, 0}            /* str */
};

#define NUM_SELECTS 4

#define SEL_TTY  0
#define SEL_UID  1
#define SEL_OWN  2
#define SEL_STAT 3

typedef struct select_info_s {
  int sel_type;
  dev_list_t tty;
  uid_list_t uid;
  char stat[6];
  string_list_t str;
  int_list_t sel;
} select_info_t, *select_info_ptr;

select_info_t selinfo = {
  0,
  {NULL, 0},
  {NULL, 0, 0},
  "\0",
  {NULL, 0, 0},
  {NULL, 0, 0}
};

static void filter_popdown_callback(Widget, XtPointer, XtPointer);
/* process user input */

static void select_popdown_callback(Widget, XtPointer, XtPointer);
/* process user input */

static XtCallbackRec accept_filter_callback[] = {
  {filter_popdown_callback, (XtPointer) 1},
  {NULL}
};

static XtCallbackRec reject_filter_callback[] = {
  {filter_popdown_callback, (XtPointer) 0},
  {NULL}
};

static XtCallbackRec accept_select_callback[] = {
  {select_popdown_callback, (XtPointer) 1},
  {NULL}
};

static XtCallbackRec reject_select_callback[] = {
  {select_popdown_callback, (XtPointer) 0},
  {NULL}
};

Widget filter_button = NULL;
/* check if filter menu has been created yet */
Widget filter_shell;
Widget filter_option[NUM_FILTERS];

Widget select_button;
Widget select_shell;
Widget select_option[NUM_SELECTS];

Widget filtsel_shell;
Widget filtsel_manager;
Widget filtsel_box;
Widget filtsel_label;
Widget filtsel_selector;

static void ready_pid_list(void)
/* compile list of pids for use in selection popup */
{
  static int fields[] = {P_PID, P_USER, P_COMMAND};
  process_info_ptr p = &procinfo;
  filter_info_ptr f = &filtinfo;
  int i;

  if(f->pid.num_allocated <= p->num_used)
    {
      f->pid.num_allocated = p->num_used + 1;
      f->pid.list = (pid_t *) 
	XtRealloc((char *) f->pid.list, f->pid.num_allocated * sizeof(pid_t));
    }
  if(f->str.num_allocated <= p->num_used)
    {
      f->str.list = (String *) XtRealloc((char *) f->str.list, 
					 (p->num_used + 1) * sizeof(String));
      for(i = f->str.num_allocated; i <= p->num_used; ++i)
	f->str.list[i] = NULL;
      f->str.num_allocated = p->num_used + 1;
    }
  for(i = 0; i < p->num_used; ++i)
    {
      f->pid.list[i] = p->table[i]->pid;
      if(f->str.list[i])
	XtFree(f->str.list[i]);
      f->str.list[i] = create_procstring(p->table[i], fields, 
					 XtNumber(fields));
    }
  f->pid.list[i] = 0;
  f->str.num_used = p->num_used;
  if(f->str.list[f->str.num_used])
    XtFree(f->str.list[f->str.num_used]);
  f->str.list[f->str.num_used] = NULL;
}

static void ready_uid_list(uid_list_ptr uid, string_list_ptr str)
/* make list of users for use in selection popup */
{
  static int fields[] = {P_USER, P_UID};
  process_info_ptr p = &procinfo;
  int i, j;

  str->num_used = 0;
  for(i = j = 0;; ++i)
    {
      if(uid->num_allocated <= str->num_used)
	uid->list = (uid_t *) 
	  XtRealloc((char *) uid->list, 
		    ++uid->num_allocated * sizeof(uid_t));
      if(str->num_allocated <= str->num_used)
	{
	  str->list = (String *)
	    XtRealloc((char *) str->list, 
		      ++str->num_allocated * sizeof(String));
	  str->list[str->num_used] = NULL;
	}
      if(i >= p->num_used)
	break;
      for(j = 0; j < str->num_used; ++j)
	if(uid->list[j] == p->table[i]->uid)
	  break;
      if(str->num_used <= j)
	{
	  uid->list[j] = p->table[i]->uid;
	  if(str->list[j])
	    XtFree(str->list[j]);
	  str->list[j] = create_procstring(p->table[i], fields, 
					  XtNumber(fields));
	  ++str->num_used;
	}
    }
  if(str->list[str->num_used])
    XtFree(str->list[str->num_used]);
  str->list[str->num_used] = NULL;
}

static void ready_tty_list(dev_list_ptr tty, string_list_ptr str)
/* prepare list of ttys for use in selection popup */
{
  static int fields[] = {P_TTYL};
  process_info_ptr p = &procinfo;
  int i, j;

  str->num_used = 0;
  for(i = 0;; ++i)
    {
      if(tty->num_allocated <= str->num_used)
	tty->list = (dev_t *) 
	  XtRealloc((char *) tty->list, 
		    ++tty->num_allocated * sizeof(dev_t));
      if(str->num_allocated <= str->num_used)
	{
	  str->list = (String *)
	    XtRealloc((char *) str->list, 
		      ++str->num_allocated * sizeof(String));
	  str->list[str->num_used] = NULL;
	}
      if(i >= p->num_used)
	break;
      if(p->table[i]->tty == -1)
	continue;
      for(j = 0; j < str->num_used; ++j)
	if(tty->list[j] == p->table[i]->tty)
	  break;
      if(str->num_used <= j)
	{
	  tty->list[j] = p->table[i]->tty;
	  if(str->list[j])
	    XtFree(str->list[j]);
	  str->list[j] = create_procstring(p->table[i], fields, 
					  XtNumber(fields));
	  ++str->num_used;
	}
    }
  tty->list[str->num_used] = 0;
  if(str->list[str->num_used])
    XtFree(str->list[str->num_used]);
  str->list[str->num_used] = NULL;
}

static Boolean use_selection(void)
/* use selected processes to compile list of PIDs for filtering */
{
  process_info_ptr p = &procinfo;
  filter_info_ptr f = &filtinfo;  
  int *selection;
  int num_selected;
  int i;

  XtVaGetValues(proc_selector, 
		XtNselectedIndex, &selection,
		XtNnumberSelected, &num_selected,
		NULL);
  if(!num_selected)
    return False;
  if(f->pid.num_allocated <= num_selected)
    { 
      f->pid.num_allocated = num_selected + 1;
      /* list of PIDs is zero terminated */
      f->pid.list = (pid_t *) 
	XtRealloc((char *) f->pid.list, f->pid.num_allocated * sizeof(pid_t));
    }
  for(i = 0; i < num_selected; ++i)
    f->pid.list[i] = p->table[selection[i]]->pid;
  f->pid.list[num_selected] = 0;
  f->str.num_used = num_selected;
  /* clear selection */
  XtVaSetValues(proc_selector, XtNselectedIndex, NULL, NULL);
  return True;
}

static void use_euid(uid_list_ptr uid)
/* use own effective user id for filtering */
{
  uid_t own_euid;
  
  own_euid = geteuid();
  if(uid->num_allocated == 0)
    uid->list = (uid_t *) XtMalloc(++uid->num_allocated * sizeof(uid_t));
  uid->list[0] = own_euid;
  uid->num_used = 1;
}

static void filter_callback(Widget w, XtPointer client, XtPointer call)
/* implements internal logic of filter activation/deactivation */
{
  int filt_type = (int) client;
  int reason = (int) call;
  filter_info_ptr f = &filtinfo;
  String *choices;

  if(reason == XstoRADIO_TURNOFF)
    return;
  if(reason == XstoTURNOFF && filt_type != FILT_ANYTTY)
    {
      XstoSetCurrent(filter_option[FILT_PID], 
		     (XtPointer) filtinfo.proc_flag[filt_type]);
      return;
    }
  XtRemoveTimeOut(refresh_timer);
  switch(filt_type)
    {
    case FILT_PID:
      read_proctable(&procinfo, PROC_FILL);
      ready_pid_list();
      choices = f->str.list;
      XtVaSetValues(filtsel_shell,  
		    XtNtitle, app_resources.sel_pid_title, 
 		    NULL); 
      break;
    case FILT_TTY:
      read_proctable(&procinfo, PROC_FILL);
      ready_tty_list(&(f->tty), &(f->str));
      choices = f->str.list;
      XtVaSetValues(filtsel_shell,  
 		    XtNtitle, app_resources.sel_tty_title, 
 		    NULL); 
      break;
    case FILT_UID:
      read_proctable(&procinfo, PROC_FILL);
      ready_uid_list(&(f->uid), &(f->str));
      choices = f->str.list;
      XtVaSetValues(filtsel_shell,  
 		    XtNtitle, app_resources.sel_uid_title, 
 		    NULL); 
      break;
    case FILT_STAT:
      choices = stat_list;
      f->str.num_used = 5;
      XtVaSetValues(filtsel_shell,  
 		    XtNtitle, app_resources.sel_stat_title, 
 		    NULL); 
      break;
    case FILT_SELECT:
      if(!use_selection())
	{
	  XtVaSetValues(filter_option[FILT_NONE], XtNstate, True, NULL);
	  /* embarras user with accusing noise */ 
	  XBell(app_display, 50);
	}
      refresh_display(NULL, &refresh_timer);
      return;
    case FILT_OWN:
      use_euid(&(f->uid));
      refresh_display(NULL, &refresh_timer);
      return;
    case FILT_ANYTTY:
      if(reason == XstoTURNON)
	XtVaSetValues(filter_option[FILT_TTY], XtNstate, False, NULL);
    default:
      /* FILT_NONE: do nothing */
      refresh_display(NULL, &refresh_timer);
      return;
    }
  XtVaSetValues(filtsel_selector,
		XtNchoicesList, choices, 
		XtNnumberChoices, f->str.num_used,
		XtNacceptCallback, accept_filter_callback,
		XtNrejectCallback, reject_filter_callback,
		NULL);
  XtPopup(filtsel_shell, XtGrabExclusive);
}

static void popup_callback(Widget w, XtPointer client, XtPointer call)
/* position popup window under pointer and configure children */
{
  Window root, child;
  int x, y, root_x, root_y;
  unsigned int mask;
 
  XQueryPointer(app_display, XtWindow(app_shell),
		&root, &child, &root_x, &root_y, &x, &y, &mask);
  XtVaSetValues(filtsel_shell, XtNx, root_x, XtNy, root_y, NULL);
}

static void filter_popdown_callback(Widget w, XtPointer client, XtPointer call)
/* process selection made in popup window */
{
  int *selection = (int *) call;
  int accept_selection = (int) client;
  filter_info_ptr f = &filtinfo;
  int i;

  XtPopdown(filtsel_shell);
  if(accept_selection)
    {
      switch((int) XstoGetCurrent(filter_option[FILT_PID]))
	{
	case PROC_PID:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    f->pid.list[i] = f->pid.list[selection[i]];
	  f->pid.list[i] = 0;
	  break;
	case PROC_UID:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    f->uid.list[i] = f->uid.list[selection[i]];
	  f->uid.num_used = i;
	  break;
	case PROC_TTY:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    f->tty.list[i] = f->tty.list[selection[i]];
	  f->tty.list[i] = 0;
	  /* turn off manually without triggering callback */
	  XtVaSetValues(filter_option[FILT_ANYTTY], XtNstate, False, NULL);
	  break;
	case PROC_STAT:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    f->stat[i] = stat_list[selection[i]][0];
	  f->stat[i] = '\0';
	  break;
	}
    }
  else
    /* don't trigger callback */
    XtVaSetValues(filter_option[FILT_NONE], XtNstate, True, NULL);
  refresh_display(NULL, &refresh_timer);
}

int filter_proc_flags(void)
/* returns flags used by read_proctable to filter the proc fs */
{
  int flags;

  if(filter_button == NULL)
    /* filter menu has not been created yet */
    return 0;
  flags = (int) XstoGetCurrent(filter_option[FILT_PID]) & ~PROC_NONE;
  flags |= (int) XstoGetCurrent(filter_option[FILT_ANYTTY]);
  return flags;
}

static Boolean tty_in_list(dev_t *list, dev_t tty)
/* search for passed tty in list */
{
  while(*list)
    if(*list++ == tty)
      return True;
  return False;
}

static Boolean uid_in_list(uid_t *list, uid_t uid, int n)
/* search for passed UID in list */
{
  while(n-- > 0)
    if(*list++ == uid)
      return True;
  return False;
}

static void compile_selection(void)
/* compile list of PIDs matching the selection criteria */
{
  process_info_ptr p = &procinfo;
  select_info_ptr s = &selinfo;
  int i, j;
  
  if(s->sel.num_allocated <= p->num_used)
    {
      s->sel.num_allocated = p->num_used + 1;
      s->sel.list = (int *)
	XtRealloc((char *) s->sel.list, s->sel.num_allocated * sizeof(int));
    }
  switch(s->sel_type)
    {
    case SEL_TTY:
      for(i = j = 0; i < p->num_used; ++i)
	if(tty_in_list(s->tty.list, p->table[i]->tty)) 
	  s->sel.list[j++] = i;
      break;
    case SEL_UID:
    case SEL_OWN:
      for(i = j = 0; i < p->num_used; ++i)
	if(uid_in_list(s->uid.list, p->table[i]->uid, s->uid.num_used)) 
	  s->sel.list[j++] = i;
      break;
    case SEL_STAT:
      for(i = j = 0; i < p->num_used; ++i)
	if(strchr(s->stat, p->table[i]->state)) 
	  s->sel.list[j++] = i;
      break;
    }
  s->sel.list[j] = XlswINVALID_INDEX;
  s->sel.num_used = j;
}

static void select_callback(Widget w, XtPointer client, XtPointer call)
/* trigger actions in response to user's choice of selection filter */
{
  select_info_ptr s = &selinfo;
  String *choices;
  
  s->sel_type = (int) client;
  XtRemoveTimeOut(refresh_timer);
  switch(s->sel_type)
    {
    case SEL_TTY:
      ready_tty_list(&(s->tty), &(s->str));
      choices = s->str.list;
      XtVaSetValues(filtsel_shell, 
		    XtNtitle, app_resources.sel_tty_title,
		    NULL);
      break;
    case SEL_UID:
      ready_uid_list(&(s->uid), &(s->str));
      choices = s->str.list;
      XtVaSetValues(filtsel_shell, 
		    XtNtitle, app_resources.sel_uid_title,
		    NULL);
      break;
    case SEL_STAT:
      choices = stat_list;
      s->str.num_used = 5;
      XtVaSetValues(filtsel_shell, 
		    XtNtitle, app_resources.sel_stat_title,
		    NULL);
      break;
    case SEL_OWN:
      use_euid(&(s->uid));
      compile_selection();
      XtVaSetValues(proc_selector, 
		    XtNselectedIndex, s->sel.list,
		    XtNnumberSelected, s->sel.num_used, 
		    NULL);
      refresh_display(NULL, &refresh_timer);
      return;
    }
  XtVaSetValues(filtsel_selector,
		XtNchoicesList, choices, 
		XtNnumberChoices, s->str.num_used,
		XtNacceptCallback, accept_select_callback,
		XtNrejectCallback, reject_select_callback,
		NULL);
  XtPopup(filtsel_shell, XtGrabExclusive);
}

static void select_popdown_callback(Widget w, XtPointer client, XtPointer call)
/* process user's selection in popup window */
{
  int *selection = (int *) call;
  int accept_selection = (int) client;
  select_info_ptr s = &selinfo;
  int i;

  XtPopdown(filtsel_shell);
  if(accept_selection)
    {
      switch(s->sel_type)
	{
	case SEL_UID:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    s->uid.list[i] = s->uid.list[selection[i]];
	  s->uid.num_used = i;
	  break;
	case SEL_TTY:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    s->tty.list[i] = s->tty.list[selection[i]];
	  s->tty.list[i] = 0;
	  break;
	case SEL_STAT:
	  for(i = 0; selection[i] != XlswINVALID_INDEX; ++i)
	    s->stat[i] = stat_list[selection[i]][0];
	  s->stat[i] = '\0';
	}
      compile_selection();
      XtVaSetValues(proc_selector, 
		    XtNselectedIndex, s->sel.list,
		    XtNnumberSelected, s->sel.num_used, 
		    NULL);
    }
  refresh_display(NULL, &refresh_timer);
}
  
void create_filtsel_popup(void)
/* create popup window allowing the user to select from a list of PIDs etc. */
{
  Arg arglist[8];
  Cardinal n = 0;

  filtsel_shell = 
    XtCreatePopupShell("filtselShell", transientShellWidgetClass, app_shell,
		       arglist, n);
  XtAddCallback(filtsel_shell, XtNpopupCallback, popup_callback, 
		(XtPointer) NULL);
  XtSetArg(arglist[n], XtNmanageSet, True); n++;
  XtSetArg(arglist[n], XtNmanageClosure, True); n++;
  XtSetArg(arglist[n], XtNmanageGrip, False); n++;
  assert(n <= XtNumber(arglist));
  filtsel_selector =
    XtCreateManagedWidget("filtselSelector", listselWidgetClass,
			  filtsel_shell, arglist, n);
}

void create_filter_menu(void)
/* create menu for selecting the filter to apply */
{
  Widget radio_group = NULL;
  Arg arglist[5];
  int i;

  filter_button =
    XtCreateManagedWidget("filterButton", menuButtonWidgetClass,
			  button_bar, NULL, 0);
  filter_shell = 
    XtCreatePopupShell("filterShell", simpleMenuWidgetClass, filter_button,
		       NULL, 0);
  XtVaSetValues(filter_button, XtNmenuName, "filterShell", NULL);
  XtSetArg(arglist[0], XtNleftMargin, 20);
  XtSetArg(arglist[1], XtNplacement, XtJustifyLeft);
  XtSetArg(arglist[2], XtNonBitmap, tick_bitmap);
  for(i = 0; i < NUM_FILTERS; ++i)
    {
      XtSetArg(arglist[3], XtNradioGroup, radio_group);
      XtSetArg(arglist[4], XtNradioData, (XtPointer) filtinfo.proc_flag[i]);
      filter_option[i] = 
	XtCreateManagedWidget(filter_option_name[i], smeToggleObjectClass, 
			      filter_shell, arglist, 5);
      XtAddCallback(filter_option[i], XtNcallback, filter_callback, 
		    (XtPointer) i);
      radio_group = filter_option[0];
    }
  /* remove option from radio group, filter can be toggled independently */
  XstoChangeRadioGroup(filter_option[FILT_ANYTTY], NULL);
}

void create_select_menu(void)
/* create menu for selection by feature */
{
  int i;

  select_button =
    XtCreateManagedWidget("selectButton", menuButtonWidgetClass,
			  button_bar, NULL, 0);
  select_shell = 
    XtCreatePopupShell("selectShell", simpleMenuWidgetClass, select_button,
		       NULL, 0);
  XtVaSetValues(select_button, XtNmenuName, "selectShell", NULL);
  for(i = 0; i < NUM_SELECTS; ++i)
    {
      select_option[i] = 
	XtCreateManagedWidget(select_option_name[i], smeBSBObjectClass, 
			      select_shell, NULL, 0);
      XtAddCallback(select_option[i], XtNcallback, select_callback, 
		    (XtPointer) i);
    }
}







