/* 	$Id: xsignal.c,v 1.6 1997/07/01 09:35:39 pirx Exp $	 */

/*
 * Author: Achim Bangert (abangert@ix.urz.uni-heidelberg.de)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/resource.h>
#include <ctype.h>
#include <assert.h>

#include <X11/cursorfont.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#ifdef XAW3D
#include <X11/Xaw3d/Form.h>
#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/SimpleMenP.h>
#include <X11/Xaw3d/Sme.h>
#include <X11/Xaw3d/SmeP.h>
#include <X11/Xaw3d/SmeBSB.h>
#include <X11/Xaw3d/SmeLine.h>
#include <X11/Xaw3d/AsciiText.h>
#include <X11/Xaw3d/TextP.h>
#else
#include <X11/Xaw/Form.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SimpleMenP.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeP.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/TextP.h>
#endif
#include "Listsel.h"

#include "global.h"
#include "xtop.h"
#include "xsignal.h"

#define NUM_SIGNALS  NSIG
#define NUM_PRIORITY 21

Widget signal_button;
Widget signal_shell;
Widget signal_option[NUM_SIGNALS];

Widget report_shell;
Widget report_manager;
Widget report_text;
Widget report_command;

static XtPopdownIDRec report_popdown_id;

Widget priority_button;
Widget priority_shell;
Widget priority_option[NUM_PRIORITY + 5];

XtTranslations button_translations;
XtTranslations shell_translations;

static void change_cursor_action(Widget, XEvent *, String *, Cardinal *);
/* action that changes the cursor in response to used button */

static void priority_notify_action(Widget, XEvent *, String *, Cardinal *);
/* notify action that distinguishes between mouse buttons */

static XtActionsRec actions[] = {
  {"ChangeCursor", change_cursor_action},
  {"PriorityNotify", priority_notify_action}
};

static String button_translation_table =
"Any<BtnDown>:  reset() PopupMenu() ChangeCursor()";

static String shell_translation_table = 
"<BtnUp>: MenuPopdown() PriorityNotify() unhighlight()"; 

static void reset_report(void)
/* clear text in report window */
{
  XawTextBlock del_block;
  int length;

  del_block.firstPos = 0;
  del_block.length = 0;
  del_block.ptr = "";
  del_block.format = XawFmt8Bit;
  XawTextDisableRedisplay(report_text);
  XtVaSetValues(report_text, XtNeditType, XawtextEdit, NULL);
  length = ((TextWidget) report_text)->text.lastPos;
  XawTextSetInsertionPoint(report_text, 0);
  XawTextReplace(report_text, 0, length, &del_block);
  XtVaSetValues(report_text, XtNeditType, XawtextRead, NULL);
  XawTextEnableRedisplay(report_text); 
}

static void append_string2report(String addendum)
/* append string to report */
{
  XawTextBlock add_block;
  int length;

  add_block.firstPos = 0;
  add_block.length = strlen(addendum);
  add_block.ptr = addendum;
  add_block.format = XawFmt8Bit;
  XawTextDisableRedisplay(report_text);
  XtVaSetValues(report_text, XtNeditType, XawtextAppend, NULL);
  length = ((TextWidget) report_text)->text.lastPos;
  XawTextReplace(report_text, length, length, &add_block);
  XawTextSetInsertionPoint(report_text, length + add_block.length);
  XtVaSetValues(report_text, XtNeditType, XawtextRead, NULL);
  XawTextEnableRedisplay(report_text); 
}

static void signal_callback(Widget w, XtPointer client, XtPointer call)
/* send signal to selected processes */
{
  int signo = (int) client;
  process_info_ptr p = &procinfo;
  int *selection;
  char buffer[1024];

  selection = XlswGetSelection(proc_selector);
  if(*selection == XlswINVALID_INDEX)
    return;
  reset_report();
  XtPopup(report_shell, XtGrabNone);
  while(*selection != XlswINVALID_INDEX)
    {
      if(kill(p->table[*selection]->pid, signo))
	{
	  XBell(app_display, 50);
	  switch(errno)
	    {
	    case EINVAL: /* invalid signal */
	      sprintf(buffer, app_resources.sig_signal_error, 
		     p->table[*selection]->pid);
	      break;
	    case ESRCH:  /* pid does not exist or belongs to zombie */
	      sprintf(buffer, app_resources.sig_process_error, 
		      p->table[*selection]->pid);
	      break;
	    case EPERM:  /* no permission to send signal */
	      sprintf(buffer, app_resources.sig_permission_error, 
		      p->table[*selection]->pid);
	    }
	}
      else
	sprintf(buffer, app_resources.sig_success, p->table[*selection]->pid);
      append_string2report(buffer);
      ++selection;
    }
}

static void priority_callback(Widget w, XtPointer client, XtPointer call)
/* reset priority of the selected processes */
{
  int priority = (int) client;
  unsigned int button = (unsigned int) call;
  process_info_ptr p = &procinfo;
  int *selection;
  Boolean relative;
  char buffer[1024];

  selection = XlswGetSelection(proc_selector);
  if(*selection == XlswINVALID_INDEX)
    return;
  relative = (priority < 0);
  if(button != Button1)
    priority = -priority;
  reset_report();
  XtPopup(report_shell, XtGrabNone);
  while(*selection != XlswINVALID_INDEX)
    {
      if(relative)
	priority = p->table[*selection]->nice - priority;
      errno = 0;
      if(setpriority(PRIO_PROCESS, p->table[*selection]->pid, priority) == -1
	 && errno)
	{
	  XBell(app_display, 50);
	  switch(errno)
	    {
	    case ESRCH:  /* pid does not exist or belongs to zombie */
	      sprintf(buffer, app_resources.prio_process_error, 
		      p->table[*selection]->pid);
	      break;
	    case EPERM:  /* no permission to change stranger's process prio */
	      sprintf(buffer, app_resources.prio_permission_error, 
		      p->table[*selection]->pid);
	      break;
	    case EACCES:  /* no permission to lower priority */
	      sprintf(buffer, app_resources.prio_access_error, 
		      p->table[*selection]->pid);
	    }
	}
      else
	sprintf(buffer, app_resources.prio_success, p->table[*selection]->pid);
      append_string2report(buffer);
      ++selection;
    }
}

static void popup_callback(Widget w, XtPointer client, XtPointer call)
/* position popup window under pointer */
{
  Window root, child;
  int x, y, root_x, root_y;
  unsigned int mask;

  XQueryPointer(XtDisplay(app_shell), XtWindow(app_shell),
		&root, &child, &root_x, &root_y, &x, &y, &mask);
  XtVaSetValues(report_shell, XtNx, root_x, XtNy, root_y, NULL);
}

static void resource2signal_menu(String resource)
/* build signal menu from application resource */
{
  char buffer[128];
  char label[32];
  int signo, count;
  int num_read, num_left;

  count = 0;
  for(num_left = strlen(resource); num_left > 0; num_left -= num_read)
    {
      num_read = -1;
      sscanf(resource, "%30[^:]:%d\n%n", label, &signo, &num_read);
      if(num_read == -1)
	{
	  String param = buffer;
	  Cardinal num_param = 1;

	  sprintf(buffer, "\"%s\" in line %d of", resource, count);
	  XtAppErrorMsg(app_context, "conversionFailed", "xtopSignalMenu", 
			"XtopError", "Conversion of entry %s signalMenu " 
			"resource failed", &param, &num_param);
	}
      if(0 < signo && signo < NUM_SIGNALS)
	{
	  sprintf(buffer, "sendSignal%d", signo);
	  signal_option[count] = 
	    XtVaCreateManagedWidget(buffer, smeBSBObjectClass, signal_shell,
				    XtNlabel, label,
				    NULL);
	  XtAddCallback(signal_option[count], XtNcallback, signal_callback, 
			(XtPointer) signo);
	  ++count;
	}
      else
	{
	  String param = buffer;
	  Cardinal num_param = 1;

	  sprintf(buffer, "%d labeled \"%s\"", signo, label);
	  XtAppWarningMsg(app_context, "unknownSignal", "xtopSignalMenu", 
			  "XtopError", "Unknown signal %s ignored",
			  &param, &num_param);
	}
      resource += num_read;
    }
}

void create_report_popup(void)
/* create popup window displaying feedback on success of system calls */
{
  Arg arglist[12];
  Cardinal n = 0;
  int text_width, text_border;
  int command_width, command_border;
  int width;

  report_shell = 
    XtCreatePopupShell("reportShell", transientShellWidgetClass, app_shell,
		       arglist, n);
  XtAddCallback(report_shell, XtNpopupCallback, popup_callback, 
		(XtPointer) NULL);
  report_manager = 
    XtCreateManagedWidget("reportManager", formWidgetClass,
			  report_shell, arglist, n);
  n = 0;
  XtSetArg(arglist[n], XtNtop, XawChainTop); n++;
  XtSetArg(arglist[n], XtNbottom, XawChainBottom); n++;
  XtSetArg(arglist[n], XtNleft, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNright, XawChainRight); n++;
  XtSetArg(arglist[n], XtNdisplayCaret, False); n++;
  XtSetArg(arglist[n], XtNwrap, XawtextWrapLine); n++;
  XtSetArg(arglist[n], XtNscrollVertical, XawtextScrollAlways); n++;
  assert(n < XtNumber(arglist));
  report_text =
    XtCreateManagedWidget("reportText", asciiTextWidgetClass,
			  report_manager, arglist, n);
  n = 0;
  XtSetArg(arglist[n], XtNfromVert, report_text); n++;
  XtSetArg(arglist[n], XtNtop, XawChainBottom); n++;
  XtSetArg(arglist[n], XtNbottom, XawChainBottom); n++;
  XtSetArg(arglist[n], XtNleft, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNright, XawChainRight); n++;
  assert(n < XtNumber(arglist));
  report_command =
    XtCreateManagedWidget("reportCommand", commandWidgetClass,
			  report_manager, arglist, n);
  report_popdown_id.shell_widget = report_shell;
  report_popdown_id.enable_widget = NULL;
  XtAddCallback(report_command, 
		XtNcallback, XtCallbackPopdown,
		(XtPointer) &report_popdown_id);
  XtVaGetValues(report_text, 
		XtNwidth, &text_width,
		XtNborderWidth, &text_border,
		NULL);
  XtVaGetValues(report_command, 
		XtNwidth, &command_width,
		XtNborderWidth, &command_border,
		NULL);
  width = MAX(text_width + text_border * 2, 
	      command_width + command_border * 2);
  XtVaSetValues(report_text, XtNwidth, width - text_border * 2, NULL);
  XtVaSetValues(report_command, XtNwidth, width - command_border * 2, NULL);
}

void create_signal_menu(void)
/* create menu for selecting the signal to be sent */
{
  signal_button =
    XtCreateManagedWidget("signalButton", menuButtonWidgetClass,
			  button_bar, NULL, 0);
  signal_shell = 
    XtCreatePopupShell("signalShell", simpleMenuWidgetClass, signal_button,
		       NULL, 0);
  XtVaSetValues(signal_button, XtNmenuName, "signalShell", NULL);
  resource2signal_menu(app_resources.signal_menu);
}

static void change_cursor_action(Widget w, XEvent *event, 
				 String *params, Cardinal *num_params)
/* action that changes the cursor in response to used button */
{
  static Cursor up_arrow = 0;
  static Cursor down_arrow = 0;
  Cursor cursor;
  unsigned int button = 0;

  if(up_arrow == 0)
    up_arrow = XCreateFontCursor(XtDisplay(app_shell), XC_sb_up_arrow);
  if(down_arrow == 0)
    down_arrow = XCreateFontCursor(XtDisplay(app_shell), XC_sb_down_arrow);
  if(event->type == ButtonPress)
    button = event->xbutton.button;
  cursor = (button == Button1) ? up_arrow : down_arrow;
  XDefineCursor(XtDisplay(app_shell), XtWindow(priority_button), cursor);
}

static void priority_notify_action(Widget w, XEvent *event, 
				   String *params, Cardinal *num_params)
/* notify action that distinguishes between pressed buttons */
{
  SimpleMenuWidget smw = (SimpleMenuWidget) w;
  SmeObject entry = smw->simple_menu.entry_set;
  unsigned int button = 0;
  
  XUndefineCursor(XtDisplay(app_shell), XtWindow(priority_button));
  if((entry == NULL) || !XtIsSensitive((Widget) entry)) 
    return;
  if(event->type == ButtonRelease)
    button = event->xbutton.button;
  XtCallCallbacks((Widget) entry, XtNcallback, (XtPointer) button); 
}

void create_priority_menu(void)
/* create menu for raising/lowering the process priority */
{
  char buffer[32];
  int i;

  XtAppAddActions(app_context, actions, XtNumber(actions));
  button_translations = XtParseTranslationTable(button_translation_table);
  shell_translations = XtParseTranslationTable(shell_translation_table);
  priority_button =
    XtCreateManagedWidget("priorityButton", menuButtonWidgetClass,
			  button_bar, NULL, 0);
  priority_shell = 
    XtCreatePopupShell("priorityShell", simpleMenuWidgetClass, 
		       priority_button, NULL, 0);
  XtVaSetValues(priority_button, XtNmenuName, "priorityShell", NULL);
  XtOverrideTranslations(priority_button, button_translations);
  XtOverrideTranslations(priority_shell, shell_translations);
  for(i = 0; i < NUM_PRIORITY; ++i)
    {
      sprintf(buffer, "to%02d", i);
      priority_option[i] = 
	XtCreateManagedWidget(buffer, smeBSBObjectClass, priority_shell,
			      NULL, 0);
      XtAddCallback(priority_option[i], XtNcallback, priority_callback, 
		    (XtPointer) i);
    }
  priority_option[i] = 
    XtCreateManagedWidget("separator", smeLineObjectClass, priority_shell,
			  NULL, 0);
  priority_option[++i] = 
    XtCreateManagedWidget("byOne", smeBSBObjectClass, priority_shell,
			  NULL, 0);
  XtAddCallback(priority_option[i], XtNcallback, priority_callback, 
		(XtPointer) -1);
  priority_option[++i] = 
    XtCreateManagedWidget("byTwo", smeBSBObjectClass, priority_shell,
			  NULL, 0);
  XtAddCallback(priority_option[i], XtNcallback, priority_callback, 
		(XtPointer) -2);
  priority_option[++i] = 
    XtCreateManagedWidget("byFive", smeBSBObjectClass, priority_shell,
			  NULL, 0);
  XtAddCallback(priority_option[i], XtNcallback, priority_callback, 
		(XtPointer) -5);
  priority_option[++i] = 
    XtCreateManagedWidget("byTen", smeBSBObjectClass, priority_shell,
			  NULL, 0);
  XtAddCallback(priority_option[i], XtNcallback, priority_callback, 
		(XtPointer) -10);
}








